// Actual starting positions
1.5 => float GRID_BLOCK_DIM_X; //1.5
2.0 => float GRID_BLOCK_DIM_Y; // 2.0
GRID_BLOCK_DIM_X * 5 => float GRID_WIDTH;
GRID_BLOCK_DIM_Y * 1 => float GRID_HEIGHT;


2 => float GLOBAL_GRID_BLOCK_DIM_X;
5 => float GLOBAL_GRID_BLOCK_DIM_Y;

GLOBAL_GRID_BLOCK_DIM_X * 5 => float GLOBAL_GRID_WIDTH;
GLOBAL_GRID_BLOCK_DIM_Y * 1 => float GLOBAL_GRID_HEIGHT;

-(GLOBAL_GRID_WIDTH / 2.0) +  (0 * GLOBAL_GRID_BLOCK_DIM_X) => float GLOBAL_X_LEFT_BOUND;
-(GLOBAL_GRID_WIDTH / 2.0) +  (5 * GLOBAL_GRID_BLOCK_DIM_X) => float GLOBAL_X_RIGHT_BOUND;
-(GLOBAL_GRID_HEIGHT / 2.0) + (1 * GLOBAL_GRID_BLOCK_DIM_Y) => float GLOBAL_Y_UPPER_BOUND;
-(GLOBAL_GRID_HEIGHT / 2.0) + (0 * GLOBAL_GRID_BLOCK_DIM_Y) => float GLOBAL_Y_LOWER_BOUND;
[ 
    @(0.0/255.0, 255.0/255.0, 255.0/255.0),
    @(0.0/255.0, 100.0/255.0, 255.0/255.0),
    @(0.0/255.0, 0.0/255.0, 255.0/255.0),
    @(90.0/255.0, 0.0/255.0, 255.0/255.0),
    @(170.0/255.0, 0.0/255.0, 255.0/255.0),
    @(255.0/255.0, 0.0/255.0, 100.0/255.0),
    @(255.0/255.0, 0.0/255.0, 0.0/255.0),
    @(255.0/255.0, 120.0/255.0, 0.0/255.0),
    @(255.0/255.0, 255.0/255.0, 0.0/255.0),
    @(120.0/255.0, 255.0/255.0, 0.0/255.0),
    @(0.0/255.0, 255.0/255.0, 0.0/255.0),
    @(0.0/255.0, 255.0/255.0, 120/255.0),

    ] @=> vec3 COLORS[];

int demo;
if(me.arg(0) == "demo") 1 => demo;
else 0 => demo;


// Audio ============================================================
512 => int WAVEFORM_LENGTH;  // openGL textures must be power of 2

Gain g => FFT fft =^ Chroma chroma => blackhole;        // mic input
10.0 => g.gain;
WAVEFORM_LENGTH * 2 => fft.size;
Windowing.hann(fft.size()) => fft.window;
// our hop size (how often to perform analysis)
fft.size()::samp => dur HOP;

GG.fullscreen();


class GPad extends GGen{
    GCircle circ --> this;
    TorusGeometry torusGeo;
    torusGeo.set(1.0, 0.01, 48, 48, 2 * Math.PI);
    FlatMaterial flatMat;
    GMesh torus --> this;
    torus.set(torusGeo, flatMat);
    circ.mat().transparent(0); // SET BACK TO 0
    torus.mat().transparent(0); // SET BACK TO 0


    float radius;
    float torusRadius;

    0 => int idx;

    fun void init(int idx, vec3 color) {
        idx => this.idx;
        flatMat.color(color);
        circ.mat().color(color);
    }

    fun void updateTorus(float sliderVal){
        2.0 * sliderVal => radius;
        (Math.exp(-Math.pow((sliderVal - 0.75), 2.0) / 0.1)) * 0.15=> torusRadius;  
        this.torusGeo.set(1.0 + radius, torusRadius, 48, 48, 2 * Math.PI);
    }
}


fun void fpsPrinter(dur d) {
    while (d => now){
        <<< GG.fps() >>>;
    }
}
spork ~ fpsPrinter(1::second);

// Graphics =============================================================

// position camera
10 => GG.camera().posZ;
GG.camera().orthographic();

// create meshes
GPad circlePads[12];
for (0 => int i; i < circlePads.size(); i++){
    circlePads[i].init(i, COLORS[i]);
    circlePads[i] --> GG.scene();
    // center position
    circlePads[i].pos(@(0, 0, 0));
    circlePads[i].sca(0.5);
    -(GRID_WIDTH / 2.0) + ((i % 6) * GRID_BLOCK_DIM_X)=> float x;
    -(GRID_HEIGHT / 2.0) + ((i > 5) * GRID_BLOCK_DIM_Y)=> float y;
    circlePads[i].pos(@(x, y, 0));
}

// let one FFT-size of time pass (to buffer)
fft.size()::samp => now;

fun float sigmoid(float x) {
    //return 1.f / (1 + Math.exp(-(50 * x) + (2.5)));
    return 1.f / (1 + Math.exp(-(30 * x) + (15)));
}

fun float max(float input[]) {
    input.size() => int n;
    input[0] => float runningMax;
    for (0 => int i; i < n; i++) {
        if (input[i] > runningMax){
            input[i] => runningMax;
        }
    }
    return runningMax;
}

fun int argmax(float input[]) {
    input.size() => int n;
    input[0] => float runningMax;
    0 => int runningIdx;
    for (0 => int i; i < n; i++) {
        if (input[i] > runningMax){
            input[i] => runningMax;
            i => runningIdx;
        }
    }
    return runningIdx;
}

fun float[] softmax(float input[]){
    input.size() => int n;
    float z[n];
    max(input) => float inputMax;
    for (0 => int i; i < n; i++) {
        input[i] - inputMax => z[i];
    }

    float numer[n];
    for (0 => int i; i < n; i++) {
        Math.exp(z[i]) => numer[i];
    }

    0.0 => float denom;
    for (0 => int i; i < n; i++) {
        numer[i] +=> denom;
    }

    float sftmx[n];
    for (0 => int i; i < n; i++) {
        numer[i] / denom +=> sftmx[i];
    }

    return sftmx;
}

fun float[] hardmax4(float input[]) {
    input.size() => int n;
    float output[n];
    max(input) => float maximum;
    argmax(input) => int maxIdx;
    for (0 => int i; i < n; i++) {
        (i == maxIdx) ? maximum : 0.0 => output[i];
    }
    return output;
}


fun float[] hardmax(float input[]) {
    input.size() => int n;
    float output[n];
    max(input) => float maximum;
    argmax(input) => int maxIdx;
    for (0 => int i; i < n; i++) {
        (i == maxIdx) ? maximum : 0.0 => output[i];
    }
    return output;
}

fun void updateChroma(int idx){
    float rawChroma;
    392.f => float freq;
    float sigEMAs[12];
    for (0 => int i; i < sigEMAs.size(); i++) {
        0.0 => sigEMAs[i];
    }
    0.90 => float beta;
    while(true){
        
        chroma.upchuck().fvals() @=> float chromaFeats[];
        /**
        cherr <= "rawChroma" <= ": ";
        for( 0 => int j; j < chroma.fvals().size(); j++ )
        {
            chroma.fval(j) => float tmp;
            cherr <= (tmp < 10 ? " ":"") <= tmp <= " ";
        }
        cherr <= IO.newline();
        **/

        //cherr <= "softChroma" <= ": ";
        softmax(chromaFeats) @=> float softChroma[];
        hardmax(chromaFeats) @=> float hardChroma[];
        hardmax4(chromaFeats) @=> float hardChroma4[];

        for( 0 => int j; j < chroma.fvals().size(); j++ )
        {
            //sigmoid(hardChroma[j]) => float sig;
            sigmoid(chroma.fval(j)) => float sig;


            (beta * sigEMAs[j]) + ((1 - beta) * sig)=> sigEMAs[j];

            circlePads[j].updateTorus(sigEMAs[j]);
            //hardChroma[j] => float tmp;
            //cherr <= (tmp < 10 ? " ":"") <= tmp <= " ";
            circlePads[j].circ.mat().alpha(Math.max(sigEMAs[j] - 0.1, 0.0)); //SET BACK
            circlePads[j].torus.mat().alpha(Math.max(sigEMAs[j] - 0.1, 0.0)); //SET BACK
        }
        //cherr <= IO.newline();
        HOP => now;
    }
} spork ~ updateChroma(0);

// HI ======================================
/**
GPlane plane;
plane.pos(@(0.0, 0.0, -1.0));
plane.sca(@(12.0, 8.0, 1.0));
plane.mat().color(@(0.0, 0.0, 0.0));
plane --> GG.scene();
**/
// setup geometry and materials

PlaneGeometry planeGeo;

ShaderMaterial shaderMat; // custom shader material
shaderMat.fragShader(
    me.dir() + "ColorFrag2.glsl"  // path to fragment shader
);
shaderMat.uniformFloat2("iResolution", GG.frameWidth(), GG.frameHeight());
shaderMat.transparent(0);

DataTexture tex;

// create meshes
GMesh mesh;

// setup meshes
mesh.set( planeGeo, shaderMat );
mesh --> GG.scene();
// center position

mesh.pos(2.0 * @(0, 0, -0.1));
mesh.sca(1.1 * @(10.0, 6.0, 1.0));

[45.0, 48.0, 52.0, 55.0] @=> float notes1[];
[48.0, 52, 55, 59] @=> float notes2[];
[41.0, 45, 48.0, 52] @=> float notes3[];
[36.0, 40, 43, 48] @=> float notes4[];
[35.0, 38, 35, 38] @=> float notes5[];
//[57.0] @=> float notes[];
notes1.size() => int NUM_NOTES;
[0.999, 1.001, 0.998, 1.002, 0.997, 1.003] @=> float detunes[];
//[1.0] @=> float detunes[];
detunes.size() => int NUM_UNISON;

SawOsc saw[NUM_NOTES * NUM_UNISON];
//Chorus chor[NUM_UNISON * NUM_NOTES];
//JCRev revs[NUM_UNISON * NUM_NOTES];
ADSR envs[NUM_UNISON];
PitShift shifts[NUM_UNISON];
LPF lows[NUM_UNISON];
Gain gs[NUM_UNISON];
Pan2 pans[NUM_UNISON];
SinOsc lfo => blackhole;
Gain output => dac;
0.25 => lfo.freq;
3000 => float baseCutoff;

for (0 => int i; i < NUM_UNISON; i++) { // iterate through each polyphonic voice
    envs[i] => shifts[i] => lows[i] => gs[i] => pans[i] => output;
    envs[i].set(0.8::second, 0.2::second, 1.0, 2.5::second);
    shifts[i].shift(detunes[i]);
    shifts[i].mix(1);
    0.2 => gs[i].gain;
    (i - (NUM_UNISON $ float / 2.0 )) / (NUM_UNISON $ float / 2.0 ) => pans[i].pan;
    
    for (0 => int j; j < NUM_NOTES; j++) { // iterate through a voice's polyphonies
        saw[(i * NUM_NOTES) + j] => envs[i] ; // i-th voice's j-th note will be mapped to i-th voice's envelope
    }
}

fun void setChords(float notes[]){
    for (0 => int i; i < NUM_UNISON; i++) {
        for (0 => int j; j < NUM_NOTES; j++) { // iterate through a voice's polyphonies
            Std.mtof(notes[j]) => saw[(i * NUM_NOTES) + j].freq; // j-th note will be mapped to i-th voice's j-th note
        }
    
    }
}

fun void runLFO(){
    while(true){
        for (0 => int i; i < lows.size(); i++) {
            (lfo.last() * 2000) + baseCutoff => lows[i].freq;
        }
        1::samp => now;
    }
    
}

fun void keysOn(){
    for (0 => int i; i < envs.size(); i++) {
        1 => envs[i].keyOn;
    }
}

fun void keysOff(){
    for (0 => int i; i < envs.size(); i++) {
        1 => envs[i].keyOff;
    }
}

"synth_guitar.wav" => string filePath;

fun void playGuitar(string filePath) {
    // Create a sound buffer
    SndBuf mySound => output;
    100.0 => mySound.gain;

    1.0::second => now;

    // Load the .wav file
    filePath => mySound.read;

    // Calculate the duration of the sound in seconds
    mySound.samples() / mySound.rate() => float soundDuration;


    // Start playing
    mySound => blackhole;

    // Wait for the duration of the sound
    soundDuration::second => now;
}

fun void randomCircleMovement(int i ){
    0.001 => float tol;
    10::ms => dur delta;

    while(true){
        vec3 xi;
        vec3 yi;
        vec3 target;

        if (Math.random2(0, 1) == 0) {
            if (Math.random2(0, 1) == 0){
                @(GLOBAL_X_LEFT_BOUND, 
                Math.random2f(GLOBAL_Y_LOWER_BOUND, GLOBAL_Y_UPPER_BOUND), 
                1) => target;
            } else {
                @(GLOBAL_X_RIGHT_BOUND, 
                Math.random2f(GLOBAL_Y_LOWER_BOUND, GLOBAL_Y_UPPER_BOUND), 
                1) => target;
            }

        } else {
            if (Math.random2(0, 1) == 0){
                @(Math.random2f(GLOBAL_X_LEFT_BOUND, GLOBAL_X_RIGHT_BOUND), 
                GLOBAL_Y_UPPER_BOUND,
                1) => target;
            } else {
                @(Math.random2f(GLOBAL_X_LEFT_BOUND, GLOBAL_X_RIGHT_BOUND), 
                GLOBAL_Y_LOWER_BOUND, 
                1) => target;
            }
        }

        //@(Math.random2f(GLOBAL_X_LEFT_BOUND, GLOBAL_X_RIGHT_BOUND), 
        //Math.random2f(GLOBAL_Y_LOWER_BOUND, GLOBAL_Y_UPPER_BOUND), 
        //1) => vec3 target;

        circlePads[i].pos() => vec3 orig_position;

        xi.set(orig_position.x, target.x, .00075 * (second/delta));
        yi.set(orig_position.y, target.y, .00075 * (second/delta));

        while(true){
            xi.interp(delta) => float newX;
            yi.interp(delta) => float newY;

            circlePads[i].pos() => vec3 oldPos;
            oldPos.x => float oldX;
            oldPos.y => float oldY;

            if ((Std.fabs( newX - oldX) < tol) && (Std.fabs( newY - oldY) < tol) ){
                break;
            }
            circlePads[i].pos(@(newX, newY, 1));
            delta => now;
        }
    }

}

fun void randomMovement(){
    for (0 => int i; i < 12; i++){
        spork ~ randomCircleMovement(i);
    }
    while(true){
        1::day => now;
    }
}

fun void playTrack(){
    4::second => now;

    for (0 => int i; i < 15; i++) {
        setChords(notes1);
        keysOn();
        2::second => now;
        keysOff();
        1.5::second => now;
        setChords(notes2);
        keysOn();
        2::second => now;
        keysOff();
        1.5::second => now;
        setChords(notes3);
        keysOn();
        2::second => now;
        keysOff();
        1.5::second => now;
        setChords(notes4);
        keysOn();
        1::second => now;
        keysOff();
        1.0::second => now;
        setChords(notes5);
        keysOn();
        1::second => now;
        keysOff();
        if (i == 1) {
            spork ~ playGuitar(filePath);
            spork ~ randomMovement();
        }
        1.0::second => now;
    }
    10::second => now;
}

fun void updateColorBackground(float elapsed_time) {
    shaderMat.uniformFloat("u_Time", elapsed_time / 2.5);
    shaderMat.alpha((Math.sin(elapsed_time)+1) / 2.0);
}


fun void moveCircles(float elapsed_time){
    0 => int idx;
    [10.0, 5.0, 4.0, 8.0, 3.0, 9.0, 
    6.0, 11.0, 1.0, 0.0, 2.0, 7.0] @=> float indices[];
    for (auto pad : circlePads){
        
        pad.circ.pos(@(0.05 * Math.sin(4*elapsed_time + (Math.PI * (indices[idx] / 12.0))), 0.05 * Math.sin(4*elapsed_time + (Math.PI * (((indices[idx] + 3) % 12) / 12.0))), 0));
        pad.torus.pos(@(0.05 * Math.sin(4*elapsed_time + (Math.PI * (indices[idx] / 12.0))), 0.05 * Math.sin(4*elapsed_time + (Math.PI * (((indices[idx] + 3) % 12) / 12.0))), 0));
        (idx + 1) % 12 => idx;
    }
}

fun void throbTorus(float elapsed_time){
    for (auto pad : circlePads){
        pad.torus.sca(1.0 + (0.02 * Math.sin(10* elapsed_time)));
    }
}

fun void throbRadius(float elapsed_time){
    for (auto pad : circlePads){
        pad.circ.sca(1.0 + (0.04 * Math.sin(15* elapsed_time)));
    }
}

fun void waitAndThenMove(){
    10::second => now;
    spork ~ randomMovement();
    while(true){
        10::second => now;
    }
}

if (demo){
    output => g;
    spork ~ runLFO();
    spork ~ playTrack();
} else {
    adc => g;
    spork ~ waitAndThenMove();
}

// Game loop  =============================================================
0.0 => float elapsed_time;


while (true) {
    GG.dt() => float dt;
    dt +=> elapsed_time;
    // Update logic
    updateColorBackground(elapsed_time);
    moveCircles(elapsed_time);
    throbTorus(elapsed_time);
    throbRadius(elapsed_time);
    // End update, begin render
    GG.nextFrame() => now;
}
